/*
  Author: Jens Hjalmarsson
  Contact: jens.hjalmarsson@gmail.com
*/
/*
  Main razzia class
*/
class Razzia{
  constructor(){
    this._parentDOMElement = null;
    this._progressiveWebApp = {manifest: null, serviceWorker: null}
  }
  pwa(manifest, script){
    if(typeof manifest != 'string' || typeof script != 'string'){
      throw new Error(`Invalid pwa arguments. Expecting strings`)
    }
    this._progressiveWebApp = {
      manifest: manifest,
      serviceWorker: script,
    }
  }
  initialize(tNode, parent){
    if(!(parent instanceof HTMLElement)){
      throw new Error("Invalid application HTMLElement target");
    }
    // if it's a progressive webapp
    if('serviceWorker' in navigator && this._progressiveWebApp.manifest && this._progressiveWebApp.serviceWorker){
      // link manifest to document
      const mf = document.createElement('link')
      mf.setAttribute('rel', 'manifest')
      mf.setAttribute('href', this._progressiveWebApp.manifest)
      document.head.appendChild(mf)

      // Register a service worker hosted at the root of the
      // site using the default scope.
      navigator.serviceWorker.register(this._progressiveWebApp.serviceWorker).then((registration) => {
        // registered
      }, (error) => {
        throw new Error(`Service worker registration failed: ${error}`);
      });
    }

    this._parentDOMElement = parent
    tNode.wiab = tNode;
    let element = tNode._create(parent);
    parent.appendChild(element);
  }
}

/*
  HTML component
*/
export class component{
  constructor(props, defaults){
    this.wiab = null;
    this._properties = {}
    this.refs = {};
    this.parentNode = null;
    this.element = null;
    this._partialReferences = {};
    this._partials = [];
    this._partialPresets = [];
    this._markup = null;
    this._events = {};
    this._partialHTML = null;
    this._sanitizeElement = document.createElement('div')

    // set initial properties
    this.setProps(props, defaults);
  }
  /*
    get partial
  */
  get partials(){
    return this._partialReferences;
  }
  /*
    set partials
  */
  set partials(arg){
    throw new Error(`Use the this.setTpl({...args}) method to preload partials.`);
  }
  setTpl(partials){
    let presets = [];
    const preset = true;
    for(const partial in partials){
      presets.push({node: partials[partial], reference: partial, preset: preset})
      partials[partial]._partialHTML = this.tpl(partials[partial], partial, preset);
    }
    this._partialPresets = presets;
  }
  /*
    Deny the ability to define properties directly on the props variable so
    it's a bit clearer that it's not allowed avoiding unnecessary behaviour.
  */
  set props(props){
    throw new Error(`Not allowed to set props this way, use the setProps method to define new properties`);
  }
  /*
    Return the internal properties value
  */
  get props(){
    return this._properties;
  }
  /*
    sanitize properties
  */
  _sanitize(obj){
    for(let key in obj){
      if(obj.hasOwnProperty(key) && (typeof obj[key] === "object")){
          this._sanitize(obj[key])
      }else{
        if(typeof obj[key] != 'boolean' && typeof obj[key] != 'number'){
          this._sanitizeElement.textContent = obj[key]
          obj[key] = this._sanitizeElement.innerHTML
        }
      }
    }
    return obj
  }
  /*
    Set local properties. This will only affect them localy
    and any redraw will reset the props unless they are passed
    on from their respective parent
  */
  setProps(props, defaults, redraw){
    Object.assign(this._properties, this._sanitize(defaults), this._sanitize(props));
    if(this.element && redraw != false){
      this._redraw(this._parse());
    }
  }
  /*
    Update props without re-render
    current component
  */
  setPropsStatic(props, defaults){
    this.setProps(props, defaults, false);
  }
  /*
    Clear all defined properties and apply
    new ones if they are defined
  */
  clearProps(props, defaults){
    this._properties = {};
    this.setProps(props, defaults);
  }
  /*
    Clear all defined properties and apply
    new ones if they are defined while not
    redrawing the component
  */
  clearPropsStatic(props, defaults){
    this._properties = {};
    this.setPropsStatic(props, defaults);
  }
  /*
    Set parent node that needs to be a valid
    node object
  */
  _setParent(parent){
    this.parentNode = parent;
  }
  /*
    As much as id like to get away from a componentDidMount
    kinda a thing i now realize why react does it
  */
  componentDidMount(){}
  componentWillMount(){}
  componentDidRedraw(){}
  componentWillRedraw(){}
  /*
    return the partial representation
  */
  toString(){
    return this._partialHTML
  }
  /*
    Create a partial inside the current node
  */
  tpl(tNode, reference, preset){
    /*
      Set the partials parent to this as the
      current node is the parent
    */
    tNode.wiab = this.wiab;
    tNode._setParent(this);
    /*
      Since the code would execute immediately we need
      to create a dummy placeholder container that will
      be replaced with the actual element at the rendering
      stage.
    */
    this._partials.push({node: tNode, reference: reference || null, preset: preset});
    /* make it accessible if reference is supplied */
    if(typeof reference === 'string'){
      this._partialReferences[reference] = tNode;
    }
    return `<partial data-id="${this._partials.length-1}"></partial>`;
  }
  /*
    non capturing event
  */
  event(name, parent){
    return this._event(name, parent, false);
  }
  /*
    Create a new event
  */
  _event(name, parent, useCapture){
    // if a function handler is provided
    if(typeof name === 'function'){
      parent = name
      name = name.name
    }
    // NOTE: bubble or not? Well tbh its not bubbling at all it's just
    // either current node or from the parent, should it bubble all the way up?
    /*
      The event will bubble up one level, but! should
      it bubble up throughout the whole node chain or
      just one level to the immediate parent?
    */
    let handleEventInParent = parent || false;
    let event;
    if(typeof handleEventInParent == 'boolean'){
      if(handleEventInParent){
        event = {
          event: this.parentNode[name],
          context: this.parentNode
        }
      }else{
        event = {
          event: this[name],
          context: this
        };
      }
    }else if(typeof handleEventInParent === 'function'){
      event = {
        event: parent,
        context: this
      }
    }else{
      throw new Error('Invalid event context')
    }
    if(!event){
      throw new Error(`Invalid event. Unable to find event for "${name}`);
    }
    this._events[name] = Object.assign(event, {useCapture});
    return name;
  }
  /*
    Bind eventlistener to the DOMElement
  */
  _addEventListener(element, attribute, fname){
    if(!this[fname]){
      throw new Error(`Invalid event binding : [${fname}] is not defined.`)
    }
    let callback = (e) => {
        this[fname].call(this, e, element);
    };
    element.addEventListener(attribute, callback);
  }
  /*
    Iterate over all children
  */
  _parseChildren(children){
    for(let i = 0; i < children.length; i++){
      /*
        Look for all partial children that would
        be handled with a special case
      */
      if(children[i].nodeName === 'PARTIAL'){
        if(children[i].attributes['data-id']){
          /*
            Replace the placeholder partial with a newly created element
            from the actual partial object
          */
          let partial = this._partials[children[i].attributes['data-id'].value];
          let htmlelement = partial.node._create();
          children[i].parentNode.insertBefore(htmlelement, children[i].nextSibling);
          /*
            Avoid duplicates on redraw with partials being preset
          */
          if(partial.reference in this.refs && !partial.preset){
            throw new Error(`Partial reference "${partial.reference}" is already defined`);
          }

          if(partial.reference){
            this.refs[partial.reference] = htmlelement;
          }
          /*
            Remove the placeholder
          */
          children[i].remove();
        }
        continue;
      }
      /*
        Child is a normal DOMElement of type 1, i.e an actual
        element and not a textnode/comment or what else it could be
      */
      if(children[i].nodeType === 1){
        for(let attribute = 0; attribute < children[i].attributes.length; attribute++){
          if(children[i].attributes[attribute].name === 'ref'){
            if(children[i].attributes[attribute].value in this.refs){
              throw new Error(`Reference already defined : ${children[i].attributes[attribute].value}`)
            }
            this.refs[children[i].attributes[attribute].value] = children[i]
          }
          /*
            Handle custom event triggers.
            Make sure events have custom prefix to
            avoid setting a normal event directly on the element
          */
          if(children[i].attributes[attribute].name.charAt(0) == ':'){
            const fname = children[i].attributes[attribute].name.substring(children[i].attributes[attribute].name.indexOf(':') + 1)
            /*
              If the attribute meets the requirements proceed
              and bind a new event on the DOMElement.
            */
            this._addEventListener(
              children[i],
              fname,
              children[i].attributes[attribute].value
            );

          }
        }
        /*
          Recursivly parse all childnodes
        */
        if(children[i].childNodes){
          this._parseChildren(children[i].childNodes);
        }
      }
    }
  }
  /*
    Parse the nodes markup and create DOMElements and
    parse it's children recursivly

    This will only return a new copy of the markup as a DOMElement
  */
  _parse(){
    /*
      Reset all events and registered partials
      so they dont follow along on a new parse
    */
    this.refs = {};
    this._events = [];
    this._partials = this._partialPresets;
    /*
      Get the html content/data for the specific node.
      Trim the edges so there are not uneccessary whitespace before and after
      since a node starts within it's first element anyways. i.e every node needs a
      a wrapper dom element.
    */
    let markup = this.render();
    if(typeof markup !== 'string'){
      throw new Error(`Invalid render return value. Expecting string got "${typeof markup}" in component "${this.constructor.name}"`);
    }
    this._markup = markup.trim();

    /*
      Create a new placeholder div that will be used to
      generate the DOMElement's coming from the nodes render function.
    */
    let placeholder = document.createElement('div');
    placeholder.innerHTML = this._markup;
    /*
      The actual node wrapper is the first element child. (all nodes need a wrapper element just like react)
    */
    let element = placeholder.firstElementChild;
    let children = element.childNodes;
    /*
      Recursivly parse all children inside the firstElementChild
    */
    this._parseChildren(placeholder.childNodes);
    return element;
  }
  /*
    Create creates a DOMElement from it's markup but also
    handles repositories that is passed on/referenced from its parent.
  */
  _create(){
    /*
      since it's just now the component is ready
      we can call our componentWillMount function
    */
    this.componentWillMount.call(this);
    /*
      Set the nodes DOMElement to one parsed from it's markup
    */
    this.element = this._parse();
    /*
      since it's just now the component is ready
      we can call our componentDidMount function
    */
    this.componentDidMount.call(this);
    return this.element;
  }
  /*
    Redraw the nodes DOMElement at it's current position
  */
  _redraw(element){
    if(this.element && this.element.parentNode){
      this.componentWillRedraw.call(this);
      this.element.parentNode.replaceChild(element, this.element)
      this.element = element
      this.componentDidRedraw.call(this);
    }
  }
  render(){
    throw new Error("Render not implemented");
  }
}

export class unsanitized extends component{
  /*
    Set local properties. This will only affect them localy
    and any redraw will reset the props unless they are passed
    on from their respective parent
  */
  setProps(props, defaults, redraw){
    Object.assign(this._properties, defaults, props);
    if(this.element && redraw != false){
      this._redraw(this._parse());
    }
  }
}

/*
  Export a new instance
*/
let RazziaInstance = new Razzia();
export default RazziaInstance;
